/* ************************************************************************** */
/*                                       _____           _                    */
/*  handle/_init.c                      |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libp7                      | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2017/01/04 15:01:48                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libp7/internals.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* ************************************************************************** */
/*  Get handler data                                                          */
/* ************************************************************************** */
/**
 *	p7_get_response:
 *	Get response from handle.
 *
 *	The handle has an unknown structure for the user.
 *
 *	@arg	handle			the handle.
 *	@return					the stream.
 */

const p7_packet_t *p7_get_response(p7_handle_t *handle)
{
	return (handle->_response);
}

/**
 *	p7_get_stream:
 *	Get stream from handle.
 *
 *	Same reason than above.
 *
 *	@arg	handle			the handle.
 *	@return					the stream.
 */

const p7_stream_t *p7_get_stream(p7_handle_t *handle)
{
	return (handle->_stream);
}

/**
 *	p7_get_info:
 *	Get server info.
 *
 *	@arg	handle			the handle.
 *	@return					the server info.
 */

const p7_server_t *p7_get_info(p7_handle_t *handle)
{
	return (&handle->_server);
}

/* ************************************************************************** */
/*  Initialization                                                            */
/* ************************************************************************** */
/**
 *	start:
 *	Start the communication.
 *
 *	@arg	h				the handle.
 *	@return					the error, if any.
 */

static int start(p7_handle_t *handle, int check)
{
	int err = 0;
	log_info("so we're active? let's do what active people do then!");

	/* send initial check */
	if (check) {
		log_info("sending initial check packet");
		if ((err = p7_send_ini_check(handle))) {
			log_fatal("couldn't send check/didn't receive answer");
			return (err);
		} else if (response.type != p7_pt_ack) {
			log_fatal("received packet wasn't ack, ask your dentist");
			return (p7_error_unknown);
		}
#if LOGLEVEL <= ll_info
	} else {
		log_info("skipping initial check, we suppose already initialized");
#endif
	}

	/* discover environment */
	log_info("checkin' up da environment");
	if ((err = p7_send_cmdsys_getinfo(handle))) {
		log_fatal("couldn't send env query/receive answer");
		return (err);
	} else if (response.type != p7_pt_ack || !response.extended) {
		log_fatal("response to env query wasn't an extended ack packet");
		return (p7_error_unknown);
	}

	/* save server */
	handle->_server = response.info;

	/* get environment based on sent hardware id */
	handle->_env = p7_get_env(handle->_server.hwid);
	log_info("environment is '%s'", handle->_env->name);

	/* no error */
	return (0);
}

/**
 *	_p7_init:
 *	Initialize libp7 for real.
 *
 *	@arg	h				the handle to create
 *	@arg	active			whether to start as active or not
 *	@arg	check			if starting as active, send initial check or not
 *	@arg	stream			the P7 stream to use.
 *	@return					the error (0 if ok)
 */

int _p7_init(p7_handle_t **h, int active, int check, p7_stream_t stream)
{
	int err = 0;

	/* allocate handle */
	log_info("allocating handle");
	err = p7_error_alloc;
	*h = NULL;
	if (!(*h = malloc(sizeof(p7_handle_t))))
		goto fail;
	log_info("handle allocated masta!");
	p7_handle_t *handle = *h;
	bzero(handle, sizeof(p7_handle_t));

	/* allocating buffers */
	log_info("danger zone: allocating buffers");
	err = p7_error_alloc;
	if (!(handle->_send_buffers[0] = malloc(MAX_PACKET_SIZE)))
		goto fail;
	if (!(handle->_send_buffers[1] = malloc(MAX_PACKET_SIZE)))
		goto fail;
	if (!(handle->_recv_buffer = malloc(MAX_PACKET_SIZE + 1)))
		goto fail;
	if (!(handle->_response = malloc(sizeof(p7_packet_t))))
		goto fail;
	if (!(handle->_pixels = malloc(sizeof(uint32_t*) * MAX_VRAM_HEIGHT
		+ sizeof(uint32_t) * MAX_VRAM_WIDTH * MAX_VRAM_HEIGHT)))
		goto fail;
	log_info("we managed it captain!");

	/* fill up pixels indexes */
	int y = MAX_VRAM_HEIGHT;
	uint32_t *base = (uint32_t*)(handle->_pixels + MAX_VRAM_HEIGHT);
	while (y--) handle->_pixels[y] = base + y * MAX_VRAM_WIDTH;

	/* initialize stream */
	handle->_stream_data = stream;
	handle->_stream = &handle->_stream_data;

	/* starting as active */
	handle->_active = active;
	/* initialize environment */
	handle->_env = NULL;
	/* initialize stupid things */
	handle->_shifted = 0;
	handle->_send_buffers_size[0] = 0;
	handle->_send_buffers_size[1] = 0;
	handle->_send_buffer_id = 0;

	/* if active */
	if (active && (err = start(handle, check)))
		goto fail;

	/* initialization went alright */
	return (0);
fail:
	p7_exit(*h, 0);
	*h = NULL;
	return (err);
}

/**
 *	p7_init:
 *	Initialize and automatically find a device.
 *
 *	@arg	handle		the handle.
 *	@arg	active		whether to start as active or not.
 *	@arg	check		if active, initialize the communication?
 *	@return				the error code (0 if ok).
 */

int p7_init(p7_handle_t **handle, int active, int check)
{
	int err;
	int tries = INIT_TRIES;
	int failed = 0;

	do /* a barrel roll */ {
		if (failed) {
			log_info("Trying again in one second.");
			sleep(1);
		}

		/* platform-specific communication methods */
#ifdef __linux__
		log_info("Looking for specific Linux drivers");
		err = p7_cinit(handle, active, check, NULL, 1);
		if (err != p7_error_nocalc)
			return (err);
#endif

		/* libusb communication methods - more or less platform-independent */
		log_info("Looking for general libusb devices");
		err = p7_usbinit(handle, active, check);
		if (err != p7_error_nocalc)
			return (err);

		/* wait a little */
		log_error("didn't find the calculator!");
		failed = 1;
	} while (--tries);

	/* no found calc, by the look of it. */
	return (p7_error_nocalc);
}

/* ************************************************************************** */
/*  De-initialization                                                         */
/* ************************************************************************** */
/**
 *	p7_exit:
 *	De-initialize libp7.
 *
 *	@arg	handle		the handle to free
 *	@arg	term		whether to terminate the communication or not
 */

void p7_exit(p7_handle_t *handle, int term)
{
	log_info("exit is called.");
	/* check if handle is already freed */
	if (!handle) return;
	log_info("and handle is still there, so let's go!");

	/* terminate communication */
	if (term && handle->_active) {
		log_info("we were active, send sending terminate packet");
		if (p7_send_term(handle)) {
			log_warn("couldn't send terminate packet, already disconnected?");
		}
		/* don't bother checking ack */
	}

	/* free buffers */
	log_info("free doz buffers, soldiers!");
	/* - raw reception packet buffer - */
	if (handle->_recv_buffer) free(handle->_recv_buffer);
	/* - raw sending packet buffers - */
	if (handle->_send_buffers[0]) free(handle->_send_buffers[0]);
	if (handle->_send_buffers[1]) free(handle->_send_buffers[1]);
	/* - decoded packet buffer - */
	if (handle->_response) free(handle->_response);
	/* - pixels - */
	if (handle->_pixels) free(handle->_pixels);

	/* close stream */
	(*handle->_stream->_close)(handle->_stream->_cookie);

	/* free handle */
	free(handle);
	log_info("handle is freeeeeeeeed :D");
}
