/* ************************************************************************** */
/*                                       _____           _                    */
/*  packet/ack.c                        |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libp7                      | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2017/01/04 15:01:48                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libp7/internals.h>
#include <string.h>

/* ************************************************************************** */
/*  Raw layout of an extended ack data                                        */
/* ************************************************************************** */
typedef struct {
	/* hardware identifier - ASCII */
	unsigned char hwid[8];
	/* processor identifier - ASCII */
	unsigned char cpuid[16];

	/* preprogrammed ROM capacity - ASCII-hex (Ko) */
	unsigned char preprog_rom_capacity[8];
	/* flash ROM capacity - ASCII-hex (Ko) */
	unsigned char flash_rom_capacity[8];
	/* RAM capacity - ASCII-hex (Ko) */
	unsigned char ram_capacity[8];

	/* preprogrammed ROM version - "xx.xx.xx" + type */
	unsigned char preprog_rom_version[16];

	/* bootcode version - "xx.xx.xx" + type */
	unsigned char bootcode_version[16];
	/* bootcode offset - ASCII-hex */
	unsigned char bootcode_offset[8];
	/* bootcode size - ASCII-hex (Ko) */
	unsigned char bootcode_size[8];

	/* OS version - "xx.xx.xx" + type */
	unsigned char os_version[16];
	/* OS offset - ASCII-hex */
	unsigned char os_offset[8];
	/* OS size - ASCII-hex (Ko) */
	unsigned char os_size[8];

	/* protocol version - "x.xx" */
	unsigned char protocol_version[4];
	/* product ID */
	unsigned char product_id[16];
	/* name set by user in SYSTEM */
	unsigned char username[16];
} packetdata_ackext_t;

/* ************************************************************************** */
/*  Utilities                                                                 */
/* ************************************************************************** */
/**
 *	version_of_string:
 *	Get version from 16-bytes char buffer.
 *
 *	data is not a string as it doesn't finish with an '\0'.
 *
 *	@arg	ver			the version destination
 *	@arg	data		the raw version data ("xx.xx.xx" + type)
 */

static void version_of_string(p7_version_t *ver, const unsigned char *data)
{
	char *t;

	/* get major */
	ver->major = p7_getascii(data, 2);
	/* get minor */
	ver->minor = p7_getascii(&data[3], 2);
	/* get rev */
	ver->rev = p7_getascii(&data[6], 2);
	/* get type */
	memcpy(&ver->type, &data[8], 8);
	ver->type[8] = '\0';
	if ((t = strchr(ver->type, 0xFF))) *t = '\0';

	/* log */
	log_info("major: 0x%02x, minor: 0x%02x, rev: 0x%02x (type: '%s')",
		ver->major, ver->minor, ver->rev, ver->type);
}

/**
 *	string_of_version:
 *	Put version in 16-bytes char buffer.
 *
 *	@arg	ver		the version source.
 *	@arg	data	the raw version data destination ("xx.xx.xx" + type).
 */

static void string_of_version(unsigned char *data, const p7_version_t *ver)
{
	/* put data */
	memcpy(data, "xx.xx.xx\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF", 16);
	p7_putascii(data, ver->major, 2);
	p7_putascii(&data[3], ver->minor, 2);
	p7_putascii(&data[6], ver->rev, 2);
	memcpy(&data[8], ver->type, strlen(ver->type));
}

/* ************************************************************************** */
/*  Send ACK packets                                                          */
/* ************************************************************************** */
/**
 *	p7_send_ack:
 *	Send basic command acknowledge.
 *
 *	Default response at successful packet reception.
 *
 *	@arg	handle		the libp7 handle
 *	@arg	resp		get response or not
 *	@return				the error code (0 if ok).
 */

int p7_send_ack(p7_handle_t *handle, int resp)
{
	return (p7_send_basic(handle, p7_pt_ack, 0x00, resp));
}

/**
 *	p7_confirm_ow:
 *	Sends overwrite confirmation.
 *
 *	@arg	handle		the libp7 handle
 *	@return				the error code (0 if ok)
 */

int p7_confirm_ow(p7_handle_t *handle)
{
	/* send packet */
	return (p7_send_basic(handle, p7_pt_ack, 0x01, 1));
}

/**
 *	p7_send_eack:
 *	Sends an extended ack.
 *
 *	@arg	handle		the libp7 handle
 *	@arg	info		the libp7 server info to return.
 *	@return				the error code (0 if ok)
 */

int p7_send_eack(p7_handle_t *handle, p7_server_t *info)
{
	packetdata_ackext_t raw;

	/* put in the structure */
	memset(raw.hwid, '\xFF', 8);
	memcpy(raw.hwid, info->hwid, strnlen(info->hwid, 8));
	memset(raw.cpuid, '\xFF', 8);
	memcpy(raw.cpuid, info->cpuid, strnlen(info->cpuid, 8));

	if (info->preprog_rom_wiped) {
		memset(raw.preprog_rom_capacity, '\xFF', 8);
		memset(raw.preprog_rom_version, '\xFF', 16);
	} else {
		p7_putascii(raw.preprog_rom_capacity, info->preprog_rom_capacity, 8);
		string_of_version(raw.preprog_rom_version, &info->preprog_rom_version);
	}

	p7_putascii(raw.flash_rom_capacity, info->flash_rom_capacity, 8);
	p7_putascii(raw.ram_capacity, info->ram_capacity, 8);

	if (info->bootcode_wiped)
		memset(raw.bootcode_version, '\xFF', 32);
	else {
		string_of_version(raw.bootcode_version, &info->bootcode_version);
		p7_putascii(raw.bootcode_offset, info->bootcode_offset, 8);
		p7_putascii(raw.bootcode_size, info->bootcode_size, 8);
	}

	if (info->os_wiped)
		memset(raw.os_version, '\xFF', 32);
	else {
		string_of_version(raw.os_version, &info->os_version);
		p7_putascii(raw.os_offset, info->os_offset, 8);
		p7_putascii(raw.os_size, info->os_size, 8);
	}

	p7_putascii(raw.protocol_version, info->protocol_version.major, 1);
	raw.protocol_version[1] = '.';
	p7_putascii(&raw.protocol_version[2], info->protocol_version.minor, 2);
	memset(raw.product_id, '\xFF', 16);
	memcpy(raw.product_id, info->product_id, strnlen(info->product_id, 16));
	memset(raw.username, '\xFF', 16);
	memcpy(raw.username, info->username, strnlen(info->product_id, 16));

	/* send the packet */
	return (p7_send_ext(handle, p7_pt_ack, 0x02, &raw, sizeof(raw), 1));
}

/* ************************************************************************** */
/*  Decode an ACK packet                                                      */
/* ************************************************************************** */
/**
 *	p7_decode_ack:
 *	Get data from ack data field.
 *
 *	Layout is described in the fxReverse projet documentation.
 *
 *	@arg	ack			the ack packet destination
 *	@arg	data		the raw data
 *	@arg	data_size	the raw data size
 *	@return				if there was an error
 */

int p7_decode_ack(p7_packet_t *packet,
	const unsigned char *data, size_t data_size)
{
	/* check the data size */
	if (data_size != sizeof(packetdata_ackext_t)) return (1);

	const packetdata_ackext_t *d = (const packetdata_ackext_t*)data;
	p7_server_t *info = &packet->info;
	char *t; /* temporary char pointer */

	/* log */
	log_info("ack packet is extended");

	/* hardware identifier */
	memcpy(info->hwid, d->hwid, 8); info->hwid[8] = '\0';
	log_info("hardware identifier is '%s'", info->hwid);
	/* processor identifier */
	memcpy(info->cpuid, d->cpuid, 16); info->cpuid[16] = '\0';
	log_info("cpu identifier is '%s'", info->cpuid);

	/* preprogrammed ROM information is wiped */
	info->preprog_rom_wiped = (d->preprog_rom_version[2] == 0xFF);
#if LOGLEVEL <= ll_info
	if (info->preprog_rom_wiped)
		log_info("Preprogrammed ROM information looks wiped out!");
#endif
	/* preprogrammed ROM capacity */
	info->preprog_rom_capacity =
		p7_getascii(d->preprog_rom_capacity, 8) * 1000;
	log_info("preprogrammed ROM capacity is %" PRIuP7INT "o",
		info->preprog_rom_capacity);
	/* preprogrammed ROM version */
	log_info("preprogrammed ROM version is the following:");
	version_of_string(&info->preprog_rom_version, d->preprog_rom_version);

	/* flash ROM capacity */
	info->flash_rom_capacity = p7_getascii(d->flash_rom_capacity, 8) * 1000;
	log_info("flash ROM capacity is %" PRIuP7INT "o",
		info->flash_rom_capacity);
	/* RAM capacity */
	info->ram_capacity = p7_getascii(d->ram_capacity, 8);
	log_info("RAM capacity is %" PRIuP7INT "o", info->ram_capacity);

	/* bootcode information is wiped */
	info->bootcode_wiped = (d->bootcode_version[2] == 0xFF);
#if LOGLEVEL <= ll_info
	if (info->bootcode_wiped)
		log_info("Bootcode information looks wiped out!");
#endif
	/* bootcode version */
	log_info("bootcode version is the following:");
	version_of_string(&info->bootcode_version, d->bootcode_version);
	/* bootcode offset */
	info->bootcode_offset = p7_getascii(d->bootcode_offset, 8);
	log_info("bootcode offset is 0x%08" PRIxP7INT, info->bootcode_offset);
	/* bootcode size */
	info->bootcode_size = p7_getascii(d->bootcode_size, 8);
	log_info("bootcode size is %" PRIuP7INT "o", info->bootcode_size);

	/* OS information is wiped */
	info->os_wiped = (d->os_version[2] == 0xFF);
#if LOGLEVEL <= ll_info
	if (info->os_wiped)
		log_info("OS information looks wiped out!");
#endif
	/* OS version */
	log_info("OS version is the following:");
	version_of_string(&info->os_version, d->os_version);
	/* OS offset */
	info->os_offset = p7_getascii(d->os_offset, 8);
	log_info("OS offset is 0x%08" PRIxP7INT , info->os_offset);
	/* OS size */
	info->os_size = p7_getascii(d->os_size, 8);
	log_info("OS size is %" PRIuP7INT "o", info->os_size);

	/* protocol version */
	info->protocol_version.major = p7_getascii(d->protocol_version, 1);
	info->protocol_version.minor = p7_getascii(&d->protocol_version[2], 2);
	log_info("protocol version is %d.%02d",
		info->protocol_version.major, info->protocol_version.minor);
	/* product ID */
	memcpy(info->product_id, d->product_id, 16);
	info->product_id[16] = '\0';
	if ((t = strchr(info->product_id, 0xFF))) *t = '\0';
	log_info("product ID is %s", info->product_id);
	/* username */
	memcpy(info->username, d->username, 16); info->username[16] = '\0';
	if ((t = strchr(info->username, 0xFF))) *t = '\0';
	log_info("username is %s", info->username);

	/* no error */
	return (0);
}
