/* ************************************************************************** */
/*                                       _____           _                    */
/*  handle/file.c                       |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libp7                      | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2017/01/04 15:01:48                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libp7/internals.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

/* ************************************************************************** */
/*  Cookie structure                                                          */
/* ************************************************************************** */
typedef struct {
	FILE *_rstream;
	FILE *_wstream;
	int _rstream_close;
} file_cookie_t;

/* ************************************************************************** */
/*  Callbacks                                                                 */
/* ************************************************************************** */
/**
 *	p7_file_read:
 *	Read from a FILE.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@arg	data		the data pointer
 *	@arg	size		the data size.
 *	@arg	timeout		the timeout in milliseconds.
 *	@return				the error code (0 if ok).
 */

static int p7_file_read(void *vcookie, unsigned char *dest, size_t size,
	unsigned int timeout)
{
	(void)timeout;
	file_cookie_t *cookie = (file_cookie_t*)vcookie;

	/* main receiving loop */
	size_t recv = 0;
	do {
		/* read */
		recv = fread(dest, 1, size, cookie->_rstream);

		/* iterate */
		dest = (void*)((char*)dest + recv);
		size -= recv;

		/* check error */
		if (!recv) {
			/* approximation */
			if (errno != EAGAIN)
				break;

			/* i'm absolutely unsure about this... */
			log_info("received EAGAIN, sleep and retry");
			errno = 0;
			sleep(3);
			continue;
		}
	} while (size);

	/* check error */
	if (!recv) switch (errno) {
		/* - timeout - */
		case EINTR: /* alarm */
		case ETIMEDOUT:
#ifdef ETIME
		case ETIME:
#endif
			log_error("timeout received");
			return (p7_error_timeout);

		/* - device error - */
		case ENODEV:
		case EPIPE: case ESPIPE:
			log_error("calculator was disconnected");
			return (p7_error_nocalc);

		default:
			log_fatal("errno was %d: %s", errno, strerror(errno));
			return (p7_error_unknown);
	}

	return (0);
}

/**
 *	p7_file_write:
 *	Write to a FILE.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@arg	data		the source
 *	@arg	size		the source size
 *	@return				the libp7 error (0 if ok)
 */

static int p7_file_write(void *vcookie,
	const unsigned char *data, size_t size)
{
	file_cookie_t *cookie = (file_cookie_t*)vcookie;

	/* main sending */
	size_t sent = fwrite(data, size, 1, cookie->_wstream);

	/* check the error */
	if (!sent) switch (errno) {
		/* - timeout error - */
		case EINTR: /* alarm */
		case ETIMEDOUT:
#ifdef ETIME
		case ETIME:
#endif
			log_error("timeout received");
			return (p7_error_timeout);

		/* - device disconnected - */
		case ENODEV:
			log_fatal("calculator was disconnected");
			return (p7_error_nocalc);

		/* - unknown error - */
		default:
			log_fatal("errno was %d: %s", errno, strerror(errno));
			return (p7_error_unknown);
	}

	/* no error */
	return (0);
}

/**
 *	p7_file_close:
 *	Close a FILE cookie.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@return				the libp7 error (0 if ok)
 */

static int p7_file_close(void *vcookie)
{
	file_cookie_t *cookie = (file_cookie_t*)vcookie;
	if (cookie->_rstream_close)
		fclose(cookie->_rstream);
	return (0);
}

/* ************************************************************************** */
/*  Opening functions                                                         */
/* ************************************************************************** */
/**
 *	_p7_finit:
 *	Real finit.
 *
 *	@arg	handle			the handle to create.
 *	@arg	active			whether to start as active or not
 *	@arg	check			if starting as active, send initial check or not
 *	@arg	rstream			read filestream
 *	@arg	wstream			write filestream
 *	@arg	rstream_close	should close read stream
 *	@return					the error (0 if ok).
 */

int _p7_finit(p7_handle_t **handle, int active, int check,
	FILE *rstream, FILE *wstream, int rstream_close)
{
	/* allocate the cookie */
	file_cookie_t *cookie = malloc(sizeof(file_cookie_t));
	if (!cookie) return (p7_error_alloc);
	*cookie = (file_cookie_t){
		._rstream = rstream,
		._wstream = wstream,
		._rstream_close = rstream_close
	};

	/* initialize with dat stream */
	return (_p7_init(handle, active, check, (p7_stream_t){
		._cookie = cookie,
		._read = p7_file_read,
		._write = p7_file_write,
		._close = p7_file_close,
	}));
}

/**
 *	p7_finit:
 *	Initialize libp7 with streams.
 *
 *	@arg	handle			the handle to create
 *	@arg	active			whether to start as active or not
 *	@arg	check			if starting as active, send initial check or not
 *	@arg	rstream			read filestream
 *	@arg	wstream			write filestream
 *	@return					the error (0 if ok)
 */

int p7_finit(p7_handle_t **handle, int active, int check,
	FILE *rstream, FILE *wstream)
{
	/* check things */
	if (!rstream || !wstream) {
		log_fatal("no stream, failed fopening?");
		return (p7_error_nostream);
	} else if (!__freadable(rstream)) {
		log_fatal("input stream not readable!");
		return (p7_error_noread);
	} else if (!__fwritable(wstream)) {
		log_fatal("output stream not writable!");
		return (p7_error_nowrite);
	}

	return (_p7_finit(handle, active, check, rstream, wstream, 0));
}
