/* ************************************************************************** */
/*                                       _____           _                    */
/*  handle/libusb.c                     |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libp7                      | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2017/01/04 15:01:48                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libp7/internals.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifndef no_libusb
# include <libusb.h>
/* ************************************************************************** */
/*  Cookie structure                                                          */
/* ************************************************************************** */
/* libusb */
typedef struct {
	libusb_context *_context;
	libusb_device_handle *_handle;

	unsigned char _buffer[2048];
	size_t _buffer_size;
	ssize_t _buffer_start, _buffer_end;
} libusb_cookie_t;

/* ************************************************************************** */
/*  Callbacks                                                                 */
/* ************************************************************************** */
/**
 *	p7_libusb_read:
 *	Read using libusb cookie.
 *
 *	@arg	vcookie		the cookie (voided)
 *	@arg	data		the data pointer.
 *	@arg	size		the data size.
 *	@arg	timeout		the timeout in milliseconds.
 *	@return				the error code (0 if ok).
 */

# define ENDPOINT_IN (LIBUSB_ENDPOINT_IN | LIBUSB_TRANSFER_TYPE_BULK)
static int p7_libusb_read(void *vcookie,
	unsigned char *dest, size_t size, unsigned int timeout)
{
	libusb_cookie_t *cookie = (libusb_cookie_t*)vcookie;
	/* transmit what's already in the buffer */
	if (cookie->_buffer_start <= cookie->_buffer_end) {
		size_t tocopy = cookie->_buffer_end - cookie->_buffer_start + 1;
		if (tocopy > size) tocopy = size;

		memcpy(dest, &cookie->_buffer[cookie->_buffer_start], tocopy);
		cookie->_buffer_start += tocopy;
		dest += tocopy;
		size -= tocopy;

		/* check if still needs some bytes */
		if (!size) return (0);
	}

	do {
		size_t toread_fornow = cookie->_buffer_size;
		if (toread_fornow > size) toread_fornow = size;

		/* fill buffer (main libusb interaction) */
		cookie->_buffer_start = 0;
		cookie->_buffer_end = -1;
		do {
			/* receive */
			int recv;
			int err = libusb_bulk_transfer(cookie->_handle, ENDPOINT_IN,
				&cookie->_buffer[cookie->_buffer_start], cookie->_buffer_size,
				&recv, timeout);

			/* check error */
			if (err) switch (err) {
				case LIBUSB_ERROR_PIPE:
				case LIBUSB_ERROR_NO_DEVICE:
				case LIBUSB_ERROR_IO:
					return (p7_error_nocalc);

				default:
					log_fatal("libusb error was %d: %s", err,
						libusb_strerror(err));
					return (p7_error_unknown);
			}

			/* iterate */
			cookie->_buffer_end += (ssize_t)recv;
		} while (cookie->_buffer_end < (ssize_t)toread_fornow);

		/* copy to destination */
		memcpy(dest, cookie->_buffer, toread_fornow);
		dest += toread_fornow;
		size -= toread_fornow;

		/* correct buffering */
		cookie->_buffer_start += toread_fornow;
	} while (size);

	/* no error */
	return (0);
}

/**
 *	p7_libusb_write:
 *	Write using libusb cookie.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@arg	data		the source.
 *	@arg	size		the source size.
 *	@return				the libp7 error (0 if ok).
 */

# define ENDPOINT_OUT (LIBUSB_ENDPOINT_OUT | LIBUSB_TRANSFER_TYPE_ISOCHRONOUS)
static int p7_libusb_write(void *vcookie,
	const unsigned char *data, size_t size)
{
	libusb_cookie_t *cookie = (libusb_cookie_t*)vcookie;
	/* send */
	int sent;
	int err = libusb_bulk_transfer(cookie->_handle, ENDPOINT_OUT,
		(unsigned char*)data, size, &sent, 0);

	/* check the error */
	if (err) switch (err) {
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			return (p7_error_nocalc);

		default:
			log_fatal("errno was %d: %s", errno, strerror(errno));
			return (p7_error_unknown);
	}

	/* no error! */
	return (0);
}

/**
 *	p7_libusb_close:
 *	Close libusb cookie.
 *
 *	@arg	vcookie		the cookie (uncasted)
 */

static int p7_libusb_close(void *vcookie)
{
	libusb_cookie_t *cookie = (libusb_cookie_t*)vcookie;
	libusb_close(cookie->_handle);
	libusb_exit(cookie->_context);
	free(vcookie);
	return (0);
}

/* ************************************************************************** */
/*  With libusb                                                               */
/* ************************************************************************** */
# include <stdlib.h>
# include <string.h>
# include <errno.h>

/**
 *	p7_usbinit:
 *	Initialize libp7 with USB device using libusb.
 *
 *	@arg	handle		the handle to create.
 *	@arg	active		hai.
 *	@arg	check		should we check in?
 *	@return				the error code (0 if you're a knoop).
 */

int p7_usbinit(p7_handle_t **handle, int active, int check)
{
	int err = 0;
	libusb_context *context = NULL;
	libusb_device_handle *dhandle = NULL;

	/* open up context */
	if (libusb_init(&context)) {
		log_fatal("Couldn't create libusb context.");
		return (p7_error_nocalc);
	}

	/* get device list */
	libusb_device **device_list = NULL;
	int device_count = libusb_get_device_list(context, &device_list);
	if (device_count < 0) {
		log_fatal("couldn't get device list.");
		err = p7_error_nocalc;
		goto fail;
	}

	/* look for the calculator */
	libusb_device *calc = NULL;
	for (int id = 0; id < device_count; id++) {
		/* get the device descriptor */
		struct libusb_device_descriptor descriptor;
		if (libusb_get_device_descriptor(device_list[id], &descriptor))
			continue;

		/* check the IDs */
		if (_p7_check_usb(descriptor.idVendor, descriptor.idProduct)) {
			calc = device_list[id];
			break;
		}
	}

	/* the calculator wasn't found */
	if (!calc) {
		libusb_free_device_list(device_list, 1);
		err = p7_error_nocalc;
		goto fail;
	}

	/* get calculator handle */
	log_info("getting the device handle");
	int uerr = libusb_open(calc, &dhandle);
	libusb_free_device_list(device_list, 1);

	/* check if we have the handle */
	switch (uerr) {
	/* couldn't get access to the device */
	case LIBUSB_ERROR_ACCESS:
		err = p7_error_noaccess;
		goto fail;

	/* default case */
	case 0: break;
	default:
		log_fatal("libusb_open returned %d: %s", uerr, libusb_error_name(uerr));
		log_fatal("errno was %d: %s", errno, strerror(errno));
		err = p7_error_noaccess;
		goto fail;
	}

	/* make the cookie */
	libusb_cookie_t *cookie = malloc(sizeof(libusb_cookie_t));
	if (!cookie) { free(cookie); goto fail; }
	*cookie = (libusb_cookie_t){
		._context = context,
		._handle = dhandle,
		._buffer_size = 2048,
		._buffer_start = 0,
		._buffer_end = -1
	};

	/* final call. */
	return (_p7_init(handle, active, check, (p7_stream_t){
		._cookie = cookie,
		._read = p7_libusb_read,
		._write = p7_libusb_write,
		._close = p7_libusb_close
	}));

fail:
	if (handle) libusb_close(dhandle);
	if (context) libusb_exit(context);
	return (err);
}

/* ************************************************************************** */
/*  Without libusb: placebos                                                  */
/* ************************************************************************** */
#else

/**
 *	p7_usbinit:
 *	Initialize libp7 with USB device [not] using libusb. Placebo.
 *
 *	@arg	handle		the handle to create.
 *	@arg	active		hai.
 *	@arg	check		should we check in?
 *	@return				the error code (0 if you're a knoop).
 */

int p7_usbinit(p7_handle_t **handle, int active, int check)
{
	(void)handle;
	(void)active;
	(void)check;
	return (p7_error_nocalc);
}

#endif
